import json
import os
import sqlite3
import threading
import time
import uuid


class Storage:
    def __init__(self, path=None):
        self.path = path or os.path.join(os.getcwd(), ".gilas.db")
        self._lock = threading.Lock()
        self._conn = sqlite3.connect(self.path, check_same_thread=False)
        self._conn.execute("PRAGMA journal_mode=WAL;")
        self._conn.execute("PRAGMA synchronous=NORMAL;")
        self._conn.execute(
            "CREATE TABLE IF NOT EXISTS objects ("
            "id TEXT PRIMARY KEY, "
            "type TEXT NOT NULL, "
            "data TEXT NOT NULL, "
            "updated_at REAL NOT NULL"
            ")"
        )
        self._conn.execute(
            "CREATE TABLE IF NOT EXISTS names ("
            "name TEXT PRIMARY KEY, "
            "id TEXT NOT NULL, "
            "type TEXT NOT NULL, "
            "updated_at REAL NOT NULL"
            ")"
        )
        self._conn.commit()
        self._cache = {}
        self._load_cache()

    def _load_cache(self):
        cursor = self._conn.execute("SELECT id, type, data FROM objects")
        for object_id, object_type, data in cursor.fetchall():
            self._cache[object_id] = (object_type, json.loads(data))

    def generate_id(self):
        return uuid.uuid4().hex

    def get(self, object_id):
        with self._lock:
            cached = self._cache.get(object_id)
            if cached is not None:
                return cached

            cursor = self._conn.execute(
                "SELECT type, data FROM objects WHERE id = ?", (object_id,)
            )
            row = cursor.fetchone()
            if row is None:
                return None
            object_type, data = row
            parsed = json.loads(data)
            self._cache[object_id] = (object_type, parsed)
            return object_type, parsed

    def save(self, object_id, object_type, data):
        payload = json.dumps(data)
        now = time.time()
        with self._lock:
            self._cache[object_id] = (object_type, data)
            self._conn.execute(
                "INSERT INTO objects (id, type, data, updated_at) "
                "VALUES (?, ?, ?, ?) "
                "ON CONFLICT(id) DO UPDATE SET "
                "type=excluded.type, data=excluded.data, updated_at=excluded.updated_at",
                (object_id, object_type, payload, now),
            )
            self._conn.commit()

    def get_named(self, name, object_type):
        with self._lock:
            cursor = self._conn.execute(
                "SELECT id FROM names WHERE name = ? AND type = ?",
                (name, object_type),
            )
            row = cursor.fetchone()
            if row is None:
                return None
            return row[0]

    def set_named(self, name, object_id, object_type):
        now = time.time()
        with self._lock:
            self._conn.execute(
                "INSERT INTO names (name, id, type, updated_at) "
                "VALUES (?, ?, ?, ?) "
                "ON CONFLICT(name) DO UPDATE SET "
                "id=excluded.id, type=excluded.type, updated_at=excluded.updated_at",
                (name, object_id, object_type, now),
            )
            self._conn.commit()

    def delete(self, object_id):
        with self._lock:
            self._cache.pop(object_id, None)
            self._conn.execute("DELETE FROM objects WHERE id = ?", (object_id,))
            self._conn.commit()


_STORAGE = None
_STORAGE_PATH = None


def _resolve_path(path=None):
    resolved_path = path or os.environ.get("GILAS_DB_PATH")
    return resolved_path or os.path.join(os.getcwd(), ".gilas.db")


def get_storage(path=None):
    global _STORAGE, _STORAGE_PATH
    resolved_path = _resolve_path(path)
    if _STORAGE is None or resolved_path != _STORAGE_PATH:
        _STORAGE = Storage(resolved_path)
        _STORAGE_PATH = resolved_path
    return _STORAGE



import os
import subprocess
import sys
import pytest
from gilas import pvar

def _run_list_script(db_path, values):
    script = (
        "from gilas import plist\n"
        "arr = plist('numbers')\n"
        "for v in {values}:\n"
        "    arr.append(v)\n"
        "print(list(arr))\n"
    ).format(values=values)
    env = os.environ.copy()
    env["GILAS_DB_PATH"] = str(db_path)
    result = subprocess.run(
        [sys.executable, "-c", script],
        env=env,
        check=True,
        capture_output=True,
        text=True,
    )
    return result.stdout.strip()


def _run_dict_script(db_path, payload, delete_keys=None, set_default=None):
    delete_keys = delete_keys or []
    set_default = set_default or {}
    script = (
        "from gilas import pdict\n"
        "data = pdict('settings')\n"
        "data.update({payload})\n"
        "for k, v in {set_default}.items():\n"
        "    data.setdefault(k, v)\n"
        "for k in {delete_keys}:\n"
        "    data.pop(k, None)\n"
        "print(sorted(dict(data).items()))\n"
    ).format(payload=payload, delete_keys=delete_keys, set_default=set_default)
    env = os.environ.copy()
    env["GILAS_DB_PATH"] = str(db_path)
    result = subprocess.run(
        [sys.executable, "-c", script],
        env=env,
        check=True,
        capture_output=True,
        text=True,
    )
    return result.stdout.strip()

def _run_var_script(db_path, value=None, set_value=False):
    script = (
        "from gilas import pvar\n"
        "v = pvar('counter')\n"
    )
    if set_value:
        script += "v.value = {value}\n".format(value=value)
    script += "print(v.value)\n"

    env = os.environ.copy()
    env["GILAS_DB_PATH"] = str(db_path)
    result = subprocess.run(
        [sys.executable, "-c", script],
        env=env,
        check=True,
        capture_output=True,
        text=True,
    )
    return result.stdout.strip()

def test_pvar_persistence_across_process_restart(tmp_path):
    db_path = tmp_path / "store.db"

    first = _run_var_script(db_path, value=1, set_value=True)
    second = _run_var_script(db_path)

    assert first == "1"
    assert second == "1"

def test_pvar_rejects_non_primitives(tmp_path, monkeypatch):
    db_path = tmp_path / "store.db"
    monkeypatch.setenv("GILAS_DB_PATH", str(db_path))

    with pytest.raises(TypeError):
        pvar("bad", initial=[1, 2, 3])

    v = pvar("ok", initial=1)
    with pytest.raises(TypeError):
        v.value = {"a": 1}

def test_persistence_across_process_restart(tmp_path):
    db_path = tmp_path / "store.db"

    first = _run_list_script(db_path, [1, 2, 3])
    second = _run_list_script(db_path, [4, 5, 6])

    assert first == "[1, 2, 3]"
    assert second == "[1, 2, 3, 4, 5, 6]"


def test_list_crud_across_process_restart(tmp_path):
    db_path = tmp_path / "store.db"

    _run_list_script(db_path, [1, 2, 3])
    script = (
        "from gilas import plist\n"
        "arr = plist('numbers')\n"
        "arr[1] = 20\n"
        "arr.append(4)\n"
        "arr.remove(1)\n"
        "arr.pop()\n"
        "print(list(arr))\n"
    )
    env = os.environ.copy()
    env["GILAS_DB_PATH"] = str(db_path)
    result = subprocess.run(
        [sys.executable, "-c", script],
        env=env,
        check=True,
        capture_output=True,
        text=True,
    )

    assert result.stdout.strip() == "[20, 3]"


def test_dict_persistence_across_process_restart(tmp_path):
    db_path = tmp_path / "store.db"

    first = _run_dict_script(db_path, {"alpha": "beta"})
    second = _run_dict_script(db_path, {"gamma": "delta"})

    assert first == "[('alpha', 'beta')]"
    assert second == "[('alpha', 'beta'), ('gamma', 'delta')]"


def test_dict_crud_across_process_restart(tmp_path):
    db_path = tmp_path / "store.db"

    _run_dict_script(db_path, {"alpha": "beta", "gamma": "delta"})
    updated = _run_dict_script(
        db_path,
        {"alpha": "omega", "epsilon": "zeta"},
        delete_keys=["gamma"],
        set_default={"theta": "lambda", "alpha": "ignored"},
    )

    assert updated == "[('alpha', 'omega'), ('epsilon', 'zeta'), ('theta', 'lambda')]"



from collections.abc import MutableSequence

from .storage import get_storage


class plist(MutableSequence):
    def __init__(self, name, initial=None):
        self._storage = get_storage()
        named_id = self._storage.get_named(name, "list")
        if named_id is None:
            self.id = self._storage.generate_id()
            self._data = list(initial) if initial is not None else []
            self._persist()
            self._storage.set_named(name, self.id, "list")
        else:
            self.id = named_id
            loaded = self._storage.get(named_id)
            if loaded is None:
                self._data = list(initial) if initial is not None else []
                self._persist()
            else:
                object_type, data = loaded
                if object_type != "list":
                    raise TypeError("Stored object is not a list.")
                self._data = list(data)

    @classmethod
    def load(cls, name, initial=None):
        return cls(name=name, initial=initial)

    def __len__(self):
        return len(self._data)

    def __getitem__(self, index):
        return self._data[index]

    def __setitem__(self, index, value):
        self._data[index] = value
        self._persist()

    def __delitem__(self, index):
        del self._data[index]
        self._persist()

    def insert(self, index, value):
        self._data.insert(index, value)
        self._persist()

    def __iter__(self):
        return iter(self._data)

    def __repr__(self):
        return repr(self._data)

    def __eq__(self, other):
        if isinstance(other, plist):
            return self._data == other._data
        return self._data == other

    def append(self, value):
        self._data.append(value)
        self._persist()

    def extend(self, values):
        self._data.extend(values)
        self._persist()

    def pop(self, index=-1):
        value = self._data.pop(index)
        self._persist()
        return value

    def clear(self):
        self._data.clear()
        self._persist()

    def remove(self, value):
        self._data.remove(value)
        self._persist()

    def reverse(self):
        self._data.reverse()
        self._persist()

    def sort(self, *args, **kwargs):
        self._data.sort(*args, **kwargs)
        self._persist()

    def __iadd__(self, other):
        self._data += list(other)
        self._persist()
        return self

    def __imul__(self, other):
        self._data *= other
        self._persist()
        return self

    def _persist(self):
        self._storage.save(self.id, "list", self._data)



from .storage import get_storage

class pvar:
    def __init__(self, name, initial=None):
        if not isinstance(name, str) or not name:
            raise ValueError("name must be a non-empty string.")

        if not self._is_primitive(initial):
            raise TypeError("pvar must be a primitive (int, float, bool, str, None)")

        self._storage = get_storage()
        self._name = name

        name_id = self._storage.get_named(name, "var")
        if name_id is None:
            self.id = self._storage.generate_id()
            self._value = initial
            self._persist()
            self._storage.set_named(name, self.id, "var")
        else:
            self.id = name_id
            loaded = self._storage.get(name_id)
            if loaded is None:
                self._value = initial
                self._persist()
            else:
                object_type, data = loaded
                if object_type != "var":
                    raise TypeError("Stored object is not a pvar")
                self._value = data
    @staticmethod
    def _is_primitive(value):
        return value is None or isinstance(value, (int, float, bool, str))
    @property
    def value(self):
        return self._value
    
    @value.setter
    def value(self, new_value):
        if not self._is_primitive(new_value):
            raise TypeError("value must be a primitive (int, float, bool, str, None)")
        self._value = new_value
        self._persist()
    
    def get(self):
        return self._value
    
    def set(self, new_value):
        self.value = new_value

    def __repr__(self):
        return repr(self._value)
    
    def _persist(self):
        self._storage.save(self.id, "var", self._value)
    


from collections.abc import MutableMapping

from .storage import get_storage

_MISSING = object()


class pdict(MutableMapping):
    def __init__(self, name, initial=None):
        self._storage = get_storage()
        named_id = self._storage.get_named(name, "dict")
        if named_id is None:
            self.id = self._storage.generate_id()
            self._data = dict(initial) if initial is not None else {}
            self._persist()
            self._storage.set_named(name, self.id, "dict")
        else:
            self.id = named_id
            loaded = self._storage.get(named_id)
            if loaded is None:
                self._data = dict(initial) if initial is not None else {}
                self._persist()
            else:
                object_type, data = loaded
                if object_type != "dict":
                    raise TypeError("Stored object is not a dict.")
                self._data = dict(data)

    @classmethod
    def load(cls, name, initial=None):
        return cls(name=name, initial=initial)

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value
        self._persist()

    def __delitem__(self, key):
        del self._data[key]
        self._persist()

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return repr(self._data)

    def __eq__(self, other):
        if isinstance(other, pdict):
            return self._data == other._data
        return self._data == other

    def clear(self):
        self._data.clear()
        self._persist()

    def pop(self, key, default=_MISSING):
        if default is _MISSING:
            value = self._data.pop(key)
        else:
            value = self._data.pop(key, default)
        self._persist()
        return value

    def popitem(self):
        item = self._data.popitem()
        self._persist()
        return item

    def setdefault(self, key, default=None):
        if key in self._data:
            return self._data[key]
        self._data[key] = default
        self._persist()
        return default

    def update(self, *args, **kwargs):
        self._data.update(*args, **kwargs)
        self._persist()

    def _persist(self):
        self._storage.save(self.id, "dict", self._data)
